############################################
# Purpose:
# This script is indented to identify which variables are used in calls to external subprogram.
#
# Inputs:
#
#    -s, --source : Path to the preprocessed sources.
#    -v, --vault : Path to the vault.
#    -o, --output : Output filename for the list of variables used in external calls.
#
############################################
from AutoRPE.UtilsRPE.SourceManager import expand_sources, load_vault, store_variable_list
from AutoRPE.UtilsRPE.BasicStructures import SubRoutine, Interface
from AutoRPE.UtilsRPE.CallManager import call_to_subroutine, find_called_subroutine, find_call_arguments
from AutoRPE.UtilsRPE.Getter import get_variable
from AutoRPE.UtilsRPE.Preprocessor import Preprocessor
import AutoRPE.UtilsRPE.CurrentBlock as CurrentBlock
from AutoRPE.UtilsRPE.Error import ProcedureNotFound
import AutoRPE.UtilsRPE.VariablePrecision as VariablePrecision
from os.path import basename, isfile, isdir


def get_command_line_arguments():
    import argparse

    parser = argparse.ArgumentParser()

    parser.add_argument("-s", "--source",
                        help="Path to the source that will be modified.", default=None)
    parser.add_argument("-v", "--vault",
                        help="Path to the vault.", default=None, metavar="FILE")
    parser.add_argument("-o", "--output",
                        help="Output filename", default=None, metavar="FILE")

    args = parser.parse_args()

    if args.source_path is None:
        parser.error("Argument -s/--source required!")
    if args.vault_path is not None:
        parser.error("Argument -v/--vault required!")
    if not isdir(args.source_path):
        parser.error("Argument -s/--source  %s is not a valid directory." % args.source)
    if not isfile(args.vault):
        parser.error("Argument -v/--vault %s is not a file." % args.vault)

    return args


def find_variables_used_in_external_calls(files, vault, exceptions):
    external_subprogram_found = []
    variables = []
    for filepath in files:
        module_name = basename(filepath).split(".")[0]
        module = vault[module_name]
        if module is None:
            continue
        file_info = Preprocessor(filepath)
        file_info.pre_process()

        current_block = CurrentBlock.CurrentBlock(module.main)
        for index, line in enumerate(file_info.lines):
            print("\r %04i/%04i %50s" % (index + 1, len(file_info.lines), module_name), end="")
            try:
                subroutine_calls = call_to_subroutine(line)
            except ValueError:
                continue
            if subroutine_calls:
                for call in subroutine_calls:
                    is_external = False
                    called_subroutine = find_called_subroutine(call)

                    # Find the procedure called (object)
                    try:
                        subroutine = vault.get_procedure_by_name(called_subroutine, current_block.module_name)
                    except ProcedureNotFound:
                        subroutine = None
                        pass

                    if isinstance(subroutine, SubRoutine):
                        continue
                    elif isinstance(subroutine, Interface):
                        continue
                    else:
                        is_external = True
                    if is_external:
                        external_subprogram_found.append(called_subroutine)
                        # Starting with the exceptions!"
                        is_exception = False
                        for exception in exceptions:
                            if called_subroutine.count(exception):
                                is_exception = True
                                break
                        if is_exception:
                            continue
                        called_arguments = find_call_arguments(call)
                        if not called_arguments:
                            # If there's no argument, continue
                            continue
                        else:
                            for argument in called_arguments:
                                try:
                                    variable = get_variable(argument, procedure=current_block.procedure, vault=vault)
                                except AssertionError:
                                    continue
                                except ValueError:
                                    continue
                                if variable is not None:
                                    if variable.type in VariablePrecision.real_types:
                                        variables.append(variable.id)
            else:
                current_block.in_which_block(line, vault=vault)
    variables = list(set(variables))
    return variables


if __name__ == "__main__":
    args = get_command_line_arguments()
    source_path = args.source
    vault_path = args.vault
    output_filename = args.output

    exceptions = ["mpi_", "xios_", "agrif_", "cice", "ice_", "wheneq", "trc_", "flush"]

    vault = load_vault(vault_path)

    files = expand_sources(source_path)

    variables = find_variables_used_in_external_calls(files, vault, exceptions)

    store_variable_list(variables, output_filename, vault)
