# Script that automatically creates different versions of a routine with different precisions
# It creates all the possible combinations of the real variables.
#
# If the subroutine has the reals foo and bar, it would generates the routines:
#   - subroutine_name_0000
#       * foo(dp)
#       * bar(dp)
#   - subroutine_name_0001
#       * foo(sp)
#       * bar(dp)
#   - subroutine_name_0002
#       * foo(dp)
#       * bar(sp)
#   - subroutine_name_0003
#       * foo(sp)
#       * bar(sp)
#

from AutoRPE.UtilsRPE.Inserter import replace_real_declaration
import AutoRPE.UtilsRPE.BasicFunctions as BasicFunctions
import AutoRPE.UtilsRPE.Vault as Vault
import AutoRPE.UtilsRPE.RegexPattern as RegexPattern

from os import mkdir
from os.path import join, isdir, abspath
import re
import itertools


def get_command_line_arguments():
    from argparse import ArgumentParser

    help_text = """
    This script takes as arguments the input source (-s/--input-source)
    and the name of a routine (-r/--routine) and creates different versions
    of this routine accounting for all different precision combinations that can happen.
    An additional argument for the vault (-v/--vault) allows to skip the
    creation of the vault if this is provided. Optional output (-o/--output-folder) for creating
    the files there, if not given the outputs are printed out.
    """
    parser = ArgumentParser(usage=help_text)

    parser.add_argument("-v", "--vault-path", required=True,
                        help="Path to a vault generated with MakeVault")
    parser.add_argument("-i", "--input-sources", required=True,
                        help="Path to the processed sources, i.e. sources with comments")
    parser.add_argument("-s", "--subprogram-name", required=True,
                        help="Name of the subroutine/function to create different interfaces")
    parser.add_argument("-o", "--output-folder", default="./",
                        help="Path to folder in which new code will be stored")
    args = parser.parse_args()

    return args


def obtain_subprogram_source(subprogram, sources_path):
    source_file = join(sources_path, subprogram.module.filename)
    indices = []
    with open(source_file) as f:
        pattern = RegexPattern.subprogram_name_declaration % subprogram.name
        lines = [l for l in f]

        for index, line in enumerate(lines):
            line = BasicFunctions.remove_comments(line).strip()
            # if it's the start or end of the subprogram
            if re.search(pattern, line, re.I):
                indices.append(index)
                if line.lower().count("end"):
                    break
        subprogram_code = lines[indices[0]:indices[1] + 1]
    for index, line in enumerate(subprogram_code):
        if line.count("::") and line.count("&"):
            subprogram_code[index] = BasicFunctions.merge_line_range(subprogram_code, index)
            for i in range(BasicFunctions.merge_line_range.number-1):
                # Remove the merged line from the list
                subprogram_code.pop(index + 1)
    return subprogram_code


def generate_routine_versions(precision_combinations, sensitive_variables, subprogram, subprogram_code, output_folder):
    code_versions = []
    for c_index, combination in enumerate(precision_combinations):
        new_subprogram_name = "%s_%04i" % (subprogram.name, c_index)
        new_subprogram_code = subprogram_code[:]
        # Search and update real declaration
        for index, variable in enumerate(sensitive_variables):
            # Find the declaration lines
            line_index = \
                [idx for idx, l in enumerate(new_subprogram_code) if re.search(r".*::.*\b%s\b" % variable.name, l, re.I)][0]
            # Update line
            new_subprogram_code[line_index] = replace_real_declaration(variable,
                                                                       new_subprogram_code[line_index],
                                                                       combination[index])

        # Update name
        new_subprogram_code[0] = new_subprogram_code[0].replace(subprogram.name, new_subprogram_name)
        new_subprogram_code[-1] = new_subprogram_code[-1].replace(subprogram.name, new_subprogram_name)
        code_versions.append("".join(new_subprogram_code))

    if not isdir(output_folder):
        mkdir(output_folder)
    # Create a file for each subroutine interface
    with open(output_folder + subprogram.name + "_interfaces.f90", 'w') as f:
        for versions in code_versions:
            f.write(versions)
            f.write("\n\n")
    print("\nAlternative procedure for %s have been written to %s" % (subprogram.name, abspath(output_folder)))


if __name__ == "__main__":

    # Get command line arguments
    args = get_command_line_arguments()
    vault_path = args.vault_path
    path_to_input_sources = args.input_sources
    subprogram_name = args.subprogram_name
    output_folder = args.output_folder

    # Load vault and retrive the object
    vault = Vault.load_vault(vault_path)
    subprogram = vault.get_procedure_by_name(subprogram_name)

    # Get the subprogram code
    subprogram_code = obtain_subprogram_source(subprogram, path_to_input_sources)

    # All possible combinations of precisions x sensitive variables
    sensitive_variables = [v for v in subprogram.dummy_arguments and v.type == "dp"]
    precision_combinations = list(itertools.product(["dp", "sp"], repeat=len(sensitive_variables)))

    # Generate the different routine versions
    generate_routine_versions(precision_combinations, sensitive_variables, subprogram, subprogram_code, output_folder)

