import AutoRPE.UtilsRPE.BasicFunctions as BasicFunctions
import AutoRPE.UtilsRPE.BasicStructures as BasicStructures
import AutoRPE.UtilsRPE.SourceManager as SourceManager
import AutoRPE.UtilsRPE.CallManager as CallManager
import AutoRPE.UtilsRPE.NatureDeterminer as NatureDeterminer
import AutoRPE.UtilsRPE.Splitter as Splitter
import AutoRPE.UtilsRPE.Getter as Getter
import AutoRPE.UtilsRPE.Error as Error
import AutoRPE.UtilsRPE.AttributeParser as AttributeParser
import AutoRPE.UtilsRPE.VariablePrecision as VariablePrecision
import warnings
import re
import itertools
from os.path import join, isfile


def insert_use_rpe_statement(vault):
    # Adding module rp_emulator (in theory adding this to par_kind.f90 should be enough,
    # but theory fails and compiler complains so I'll add it everywhere
    for module in vault.modules:
        module.lines.insert(1, "USE rp_emulator")
        module.rebuild_text()


def insert_read_precisions_module_to_file_and_save(path_to_output, vault):
    # Adding the use statement in all the files.
    line_to_add = "   USE read_precisions\n"
    for module in vault.modules:
        module.lines.insert(1, line_to_add)
        module.rebuild_text()
        module.save_modified_file(path_to_output)


def create_read_precisions_module(path_to_sources, vault=None, target_module=None):
    from os.path import dirname, abspath
    print("\n\nCreating module to read precision namelist...")

    # Create module
    rpe_vars = [var for var in vault.variables if var.type == "rpe_var"]
    assigned_rpe_vars = [var for var in rpe_vars if var.mutable]
    ids = [var.id for var in assigned_rpe_vars]
    # Path to template is hardcoded
    package_directory = dirname(abspath(__file__))
    filename = package_directory + "/../AdditionalFiles/read_precisions_template"

    with open(filename, "r") as input_file:
        template = input_file.read()
    module = template.replace("%NUMBER_OF_RPE_VARS%", "%i" % max(ids))
    with open(join(path_to_sources, "read_precisions.f90"), "w") as output_file:
        output_file.write(module)

    # Put call into target module
    if target_module:
        if isfile(join(path_to_sources, target_module)):
            with open(join(path_to_sources, target_module), "r") as input_file:
                original_source = input_file.read()

            lines = original_source.split("\n")
            lines.insert(1, "use read_precisions")
            for index, line in enumerate(lines):
                if line.lower().strip() == "private":
                    lines.pop(index)
            original_source = "\n".join(lines)

            with open(join(path_to_sources, target_module), "w") as output_file:
                output_file.write(original_source)
        else:
            warnings.warn("File %s not found in sources.\nUnable to insert read_precisions module")


def add_sbits_to_rpe_variables(vault):
    print("\n\nAssigning significant bits...")

    def condition(v):
        return v.type == "rpe_var" and not v.is_parameter and not v.name.count("tmp_var_")

    rpe_variables = [var for var in vault.variables if condition(var) and var.dimension is not None]
    allocatable, non_allocatable = Splitter.split_vars_by_allocatable(rpe_variables)
    var_counter = Counter()
    # Allocatable
    for variable in allocatable:
        insert_variable_precision_specification_to_allocatable(variable, var_counter)

    # Non allocatable that appear in a namelist
    na_in_namelist, na_not_in_namelist = Splitter.split_by_appearance_in_namelist(non_allocatable)
    for variable in na_in_namelist:
        insert_variable_precision_specification_to_namelist_parameter(variable, var_counter, vault)
    in_main, not_in_main = Splitter.split_vars_by_main_or_not(na_not_in_namelist)

    # Variables inside routines
    for variable in not_in_main:
        if not isinstance(variable.procedure, BasicStructures.DerivedType) and variable.intent != "in":
            insert_variable_precision_to_routine_variable(variable, var_counter)

    # Adding sbits assignation and apply_truncation after routine calls
    insert_variable_precision_after_subroutine_call(vault, var_counter)


def _add_argument_truncation(line, subprogram_call, counter):
    # Check the intent and type of the arguments
    if_condition = line[0]
    call_to_subroutine = line[1]
    for s_call in subprogram_call:
        variables_to_truncate = []
        if not s_call.subprogram.is_external:
            for arg, dummy_a in zip(s_call.arguments, s_call.dummy_arguments):
                if dummy_a.intent in ["out", "inout"]:
                    # The var can not be None: since the intent is out, must be a single var
                    if arg.variable is not None:
                        if arg.variable.type == "rpe_var":
                            # If is a structure the var will be a member, not an actual var: it can't be truncated
                            if arg.variable.is_member_of:
                                variables_to_truncate.append(arg.name.replace("%val", "").strip())
                            else:
                                variables_to_truncate.append(arg.variable.name)
                            if arg.variable.id is None:
                                arg.id = counter.count
                                arg.mutable = True
                                counter.up()
                    else:
                        raise Error.IntentOutError("The dummy argument %s of procedure %s has intent %s but is used "
                                                   "like %s "
                                                   % (dummy_a.name, dummy_a.procedure.name, dummy_a.intent, arg.name))
        else:
            # If the routine is external
            for argument in s_call.arguments:
                # If the argument is not a single var, it can not be intent out/inout, to %sbits will stay constant
                if argument.name.count("%val") and not NatureDeterminer.has_operations(argument.name):
                    variables_to_truncate.append(argument.name.replace("%val", "").strip())

        for variable in variables_to_truncate:
            call_to_subroutine += "\n"
            call_to_subroutine += if_condition + "    %s%%sbits = emulator_variable_precisions(%i) \n" \
                                                     % (variable, counter.count)
            call_to_subroutine += if_condition + "    CALL apply_truncation(%s)" % variable
    return call_to_subroutine


def insert_variable_precision_specification(variable, module, insertion_line, condition="", var_name=None):
    # Add precision specification
    var_name = var_name if var_name is not None else variable.name

    # Add apply_truncation
    new_line = condition + "    CALL apply_truncation(%s)" % var_name
    module.lines.insert(insertion_line, new_line)

    # Add sbits    
    new_line = condition + "    %s%%sbits = emulator_variable_precisions(%i)" % (var_name, variable.id)
    module.lines.insert(insertion_line, new_line)

    # Add variable usage specification
    new_line = condition + "    variable_is_used(%i) = .true." % variable.id
    module.lines.insert(insertion_line, new_line)

    module.rebuild_text()


def insert_variable_precision_specification_to_allocatable(variable, counter_object):
    module_where_var_appears = variable.procedure.module.used_by
    for module in [variable.module] + module_where_var_appears:
        for line_index, line in enumerate(module.lines):
            if re.search(r"\ballocate\b\s*\(.*\b" + variable.name + r"\b.*", line, re.I):
                if variable.is_member_of:
                    # Get the arguments from ALLOCATE, avoiding any possible extra call or IF statement inline
                    arguments = CallManager.find_call_arguments(BasicFunctions.remove_if_condition(line))
                    for arg in arguments:
                        if arg.count(variable.name):
                            # Remove the possible array arguments and get only member and var (foo%var(2:3) --> foo%var)
                            var_name = arg.split(variable.name)[0] + variable.name
                            break
                else:
                    var_name = None
                variable.id = counter_object.count
                variable.mutable = True
                counter_object.up()
                insert_variable_precision_specification(variable, module, line_index + 1, var_name=var_name)



def insert_variable_precision_specification_to_namelist_parameter(variable, counter_object, vault):
    module = vault[variable.appears_in_namelist[0]]
    read_index = -1
    for index, line in enumerate(module.lines):
        pattern3 = r"read *\(.*\b%s\b" % variable.appears_in_namelist[1]
        if re.search(pattern3, line, re.I):
            read_index = index
            break
    variable.id = counter_object.count
    variable.mutable = True
    counter_object.up()
    insert_variable_precision_specification(variable, module, index + 1)


def insert_variable_precision_to_routine_variable(variable, counter_object):
    module, first_use_line = find_first_use(variable)
    if module:
        var_name = None
        variable.id = counter_object.count
        variable.mutable = True
        counter_object.up()
        _line = module.lines[first_use_line]
        if re.search(r'^\s*if\s*\(', _line, re.I):
            condition = BasicFunctions.close_brackets(_line)
        else:
            condition = ""
        if variable.is_member_of:
            var_name = BasicFunctions.remove_if_condition(_line).split(variable.name)[0] + variable.name
        insert_variable_precision_specification(variable, module, first_use_line, condition, var_name=var_name)
    else:
        warnings.warn("Variable %s not found %s in routine %s" % (
            variable.name, variable.procedure.module.name, variable.procedure.name))


def insert_variable_precision_after_subroutine_call(vault, counter):
    for module in vault.modules:
        module.update_blocks(vault)
        for _index, line in enumerate(module.lines):
            # Returns arguments of subprograms call and corresponding type and dummy
            temporal_call = BasicFunctions.split_if_condition(line)
            try:
                subroutine_calls = \
                    CallManager.get_procedure_inline(temporal_call[1], vault, module.blocks[_index],
                                                only_subroutine=True)
            except Error.ProcedureNotFound as procedure:
                if procedure.args[0][0] == "apply_truncation":
                    continue
                raise Error.ProcedureNotFound(procedure.args)
            # If something was found check is using a rpe variable
            if subroutine_calls:
                module.lines[_index] = line.replace(temporal_call[1],
                                                    _add_argument_truncation(temporal_call, subroutine_calls, counter))
        module.rebuild_text()

# End fo rpe insertion


def add_val_to_argument(argument, string):
    string = replace_variable_exact_match(argument, argument + "%val", string)
    return string


def add_rpe_real_member_to_argument(argument, string):
    if NatureDeterminer.is_an_hardcoded_array(argument):
        for arg in Splitter.split_hardcoded_array(argument):
            string = replace_variable_exact_match(arg, arg + "%val", string)
        return string
    return replace_variable_exact_match(argument, argument + "%val", string)


def add_rpe_to_argument(argument, string):
    # Check we are not casting more than once
    if not string.count("rpe(" + argument + ")") == string.count(argument):
        if NatureDeterminer.is_an_hardcoded_array(argument):
            string = add_rpe_to_array_arguments(argument)
        else:
            string = replace_variable_exact_match(argument, "rpe(" + argument + ")", string)
    return string


def add_real_to_argument(argument, string, precision=VariablePrecision.real_id["wp"]):
    # Check we are not casting more than once
    precision = ", " + precision
    if string.count("real(" + argument + precision + ")") == string.count(argument):
        return string
    if NatureDeterminer.has_operations(argument):
        return string.replace(argument, "real(" + argument + precision + ")")

    if NatureDeterminer.is_an_hardcoded_array(argument):
        new_arg = add_real_to_array_arguments(argument)

    else:
        new_arg = "real(" + argument + precision + ")"

    string = replace_variable_exact_match(argument, new_arg, string)
    return string


def add_real_to_array_arguments(argument):
    simple_array = argument.replace("(/", "")
    simple_array = simple_array.replace("/)", "")
    clean_numbers = CallManager.find_call_arguments("(" + simple_array + ")")
    # Open the array
    ret_val = "(/"
    for number in clean_numbers:
        # Cast all array numbers
        ret_val += "real(" + number + " , wp), "
    # Remove last comma and closes array
    ret_val = ret_val[:-2] + "/)"
    return ret_val


def add_rpe_to_array_arguments(argument):
    simple_array = argument.replace("(/", "")
    simple_array = simple_array.replace("/)", "")
    clean_numbers = CallManager.find_call_arguments("(" + simple_array + ")")
    # Open the array
    ret_val = "(/"
    for number in clean_numbers:
        # Cast all array numbers
        ret_val += "rpe_var( 52, " + number + " ), "
    # Remove last comma and closes array
    ret_val = ret_val[:-2] + "/)"
    return ret_val


def replace_variable_exact_match(piece_to_replace, replacement, string):
    # In case string does start with a char we'll add the \b to indicate a word border
    if re.match("\w", piece_to_replace.strip()[0]):
        piece_to_replace = "\\b" + piece_to_replace.strip()

    # Adapt the string to fulfill regex rules
    piece_to_replace = piece_to_replace.replace("+", r"\+")
    piece_to_replace = piece_to_replace.replace("-", r"\-")
    piece_to_replace = piece_to_replace.replace("*", r"\*")
    if piece_to_replace.count("("):
        piece_to_replace = piece_to_replace.replace("(", "\s*\(")
        # Avoid matching structures
        if not piece_to_replace.count("%"):
            piece_to_replace = piece_to_replace.replace(")", "\)(?!%)")
        else:
            piece_to_replace = piece_to_replace.replace(")", "\)")
    else:
        piece_to_replace = piece_to_replace + "\\b(?!%val)"
        piece_to_replace = piece_to_replace + "\\b(?! *\()"
        # Before performing the substitution remove strings to avoid substitution inside them
    mask = BasicFunctions.literal_mask(string)
    literal_part = []
    n_of_literal = -1

    # Transforms the string in a list of char (strings are immutable and cannot be modified)
    string = list(itertools.chain.from_iterable([list(piece) for piece in list(string)]))
    for index, i in enumerate(mask):
        if i:
            # A new literal is beginning, add a list to store its value
            if index == 0 or not mask[index - 1]:
                n_of_literal += 1
                literal_part.append([])
            # The char is stored for late substitution
            literal_part[n_of_literal].append(string[index])
            # The char is overwrite with a placeholder
            string[index] = "#"

    # Reconvert the list to string and substitute the fix
    string = "".join(string)
    # Check the argument is passed with the same name as the keyword
    keyword = re.search(r'(%s\s*=\s*)%s' % (piece_to_replace, piece_to_replace), string, re.I)
    if keyword:
        string = list(itertools.chain.from_iterable([list(piece) for piece in list(string)]))
        # avoid casting the keyword
        len_keyword = len(keyword.group(1))
        n_of_literal += 1
        literal_part.append([])
        for i in range(keyword.start(), keyword.start()+len_keyword):
            literal_part[n_of_literal].append(string[i])
            string[i] = "#"

        string = "".join(string)

    string = re.sub(piece_to_replace, replacement, string)

    # Substitutes one at a time the placeholder for the char
    for index, substitution in enumerate(literal_part):
        string = re.sub("#+", "".join(literal_part[index]), string, 1)

    return string


def _remove_keywords(arguments):
    """
        In case there's a keyword argument, return only the value.
        """
    patt = "[a-zA-Z_0-9]* *=(.*)"
    for index, argument in enumerate(arguments):
        m = re.match(patt, argument.strip())
        if m:
            arguments[index] = m.group(1).strip()
    return arguments


def find_call_in_original_file(subprogram, argument, argument_index, _file, target_type="dp", fix=False):
    """
        Open a file and look for a specific line that contains the subprogram and the arguments.
    """
    # TODO: Probably shouldn't use sp and wp since those are model specific.
    if target_type == "sp":
        target_type = "wp"
    # Add a cast to the proper type
    new_argument = "REAL(%s, %s)" % (argument, target_type)
    # Fix in case the argument already had a cast, in which case we'll only change the type of the cast
    if new_argument.count("REAL(REAL("):
        patt = "REAL\((.*),(.*)\)"
        m = re.search(patt, argument)
        new_argument = "REAL(%s, %s)" % (m.group(1), target_type)
    # Fix the pattern to account for regex rules
    pattern = r"\b%s\b####$$$$%s" % (subprogram, argument)
    pattern = pattern.replace("(", "\(")
    pattern = pattern.replace(")", "\)")
    pattern = pattern.replace("+", "\+")
    pattern = pattern.replace("*", "\*")
    pattern = pattern.replace("####", ".*")
    pattern = pattern.replace("$$$$", r"(\b|, *)")
    pattern += r"(\b|[, )])"

    # Initialize found to False
    found = False

    # Open file
    with open(_file, "r") as f:
        lines = [l for l in f]
        for line_index, line in enumerate(lines):
            if line.lower().count("subroutine") or line.lower().count("function"):
                continue

            # In case there's a call to the target routine with a line continuation, fix it and call again the function
            if line.lower().count(subprogram):
                # The line contains the subprogram name but the pattern doesn't match
                line = BasicFunctions.remove_comments(lines[line_index]).strip()
                # Check if it is a call with split args
                if line and line.strip()[-1] == "&":
                    line = BasicFunctions.merge_line_range(lines, line_index)
                    m = re.search(pattern, line, re.I)
                    if m:
                        # It merge the split call in one single line
                        BasicFunctions.merge_line_in_file(_file, line_index)
                        # Now the function can properly fix the call
                        return find_call_in_original_file(subprogram, argument, argument_index, _file, target_type, fix)
            else:
                continue

            # Look for the call to the routine with the proper arguments
            m = re.search(pattern, line, re.I)
            if m:
                calls = CallManager.call_to_subroutine(line)

                if len(calls) > 1:
                    print("Now what?")
                arguments = _remove_keywords(CallManager.find_call_arguments(calls[0]))
                # Check if the argument is placed in the proper place
                try:
                    if arguments.index(argument) != argument_index:
                        continue
                except ValueError:
                    # It seems that this can happen if the exact same combination
                    # of routine argument has been fixed before.
                    continue
                found = True
                # TODO: Need to account for the possibility of having the same argument in different positions.
                new_line = replace_variable_exact_match(argument, new_argument, line)
                lines[line_index] = new_line + "\n"

    # In case we found the call overwrite the file with the fix
    # Otherwise return False
    if found:
        if fix:
            code = "".join(lines).strip()
            with open(_file, "w") as output:
                output.write(code)
        return True
    else:
        return False


# This function will find the file where a variable is declared and will replace the declaration
# with a new one using the new precision spec.
def replace_declaration(variable, file_paths, vault, target_precision="dp"):
    files = file_paths
    module = variable.procedure.module.name
    module_file_paths = SourceManager.match_module_and_filename(module, files)
    replacement_done = False
    for filepath in module_file_paths:
        try:
            additional_files = []
            declaration_line_indexes = find_declaration_line(variable, filepath, vault, additional_files)
        except Error.LineNotFound:
            if len(additional_files):
                for h90_file in additional_files:
                    h90_file_path = [file for file in file_paths if file.count(h90_file)]
                    module_file_paths.extend(h90_file_path)
            else:
                pass
            continue
        replacement_done = True
        with open(filepath, "r") as f:
            try:
                data = [l for l in f]
            except UnicodeDecodeError:
                raise UnicodeDecodeError("The file %s contains non UTF-8 character.\n "
                                         "Please clean, and retry.", filepath)
            for declaration_line_index in declaration_line_indexes:
                declaration_line = data[declaration_line_index]
                new_line = replace_real_declaration(variable, declaration_line, target_precision=target_precision)
                data[declaration_line_index] = new_line
            updated_text = "".join(data)
        with open(filepath, "w") as output:
            output.write(updated_text)
    if not replacement_done:
        print("Declaration of variable %s with id %i not found in module %s" % (
            variable.name, variable.id, variable.procedure.module.name))


def change_code(list_filename, source_path, vault_path, target_precision="dp"):
    # The purpose of this function is to modify the declarations of several variables in a source-code.
    # As inputs it takes the path to the file with the list of variables, the path to the sources that will be modified
    # and the path to the vault database containing information about the code.
    #
    # This function takes the following inputs:
    # - list_filename: Path to a file that contains a list of variables that will be modified in the code
    # - source_path: Path to a folder that contains the source files of the code corresponding to the vault.
    # - vault_path: Path to a vault database.

    vault = SourceManager.load_vault(vault_path)
    variable_ids = SourceManager.load_variable_list(list_filename)
    file_paths = SourceManager.expand_sources(source_path=source_path)

    variables = [vault.get_variable_by_id(_id) for _id in variable_ids]
    for variable in variables:
        replace_declaration(variable, file_paths, vault, target_precision=target_precision)


# This function takes a declaration line (string) and updates it with the new type
# If the declaration contains more variables it keeps the other variables with the original attributes.
def replace_real_declaration(variable, declaration_line, target_precision="dp"):
    declaration_line_without_comments = BasicFunctions.remove_comments(declaration_line).strip()

    attributes, arguments = declaration_line_without_comments.split("::")
    variable_name = AttributeParser.get_variable_name(arguments)

    # Real statement pattern
    pattern = "real *\(([^\)]*)\)"

    if len(variable_name) == 1:
        m = re.search(pattern, declaration_line_without_comments, re.I)
        if m:
            real_statement = m.group(0)
            precision_specification = m.group(1)
            new_real_statement = real_statement.replace(precision_specification, target_precision)
            new_declaration_line_without_comments = \
                declaration_line_without_comments.replace(real_statement, new_real_statement)
            new_declaration_line = \
                declaration_line.replace(declaration_line_without_comments, new_declaration_line_without_comments)
        else:
            warnings.warn("Problems finding real statement:\n%s" % declaration_line)
            new_declaration_line = "! WARNING!! the declaration hasn't been properly fixed!\n" + \
                                   declaration_line
        return new_declaration_line
    else:
        m = re.search(pattern, attributes, re.I)
        if m:
            real_statement = m.group(0)
            precision_specification = m.group(1)
            new_real_statement = real_statement.replace(precision_specification, target_precision)
            new_attributes = attributes.replace(real_statement, new_real_statement)
            additional_declaration = "%s :: %s" % (new_attributes, variable.name)
            m = re.search(r"(.*\b)%s(\b.*)" % variable.name, declaration_line_without_comments)
            if m:

                declaration_line_without_selected_var = "".join(m.groups(0))
            else:
                warnings.warn("WTF?")

            # Comma patterns to remove the extra comma
            beginning_pat = "::( *\,)"
            beginning_pat_substitution = "::"
            middle_pat = ",( *,)"
            middle_pat_sub = ","
            last_pat = "(, *)$"
            last_pat_sub = ""

            for pat, sub in [(beginning_pat, beginning_pat_substitution), (middle_pat, middle_pat_sub),
                             (last_pat, last_pat_sub)]:
                m = re.search(pat, declaration_line_without_selected_var)
                if m:
                    new_declaration_line_without_comments = re.sub(pat, sub, declaration_line_without_selected_var)
                    new_declaration_line = declaration_line.replace(declaration_line_without_comments,
                                                                    new_declaration_line_without_comments)
                    break
        else:
            pass

        # Adding proper indentation
        leading_spaces = (len(new_declaration_line) - len(new_declaration_line.lstrip()))
        new_declaration_line = "%s%s\n%s" % (
            leading_spaces * " ", additional_declaration.lstrip(), new_declaration_line)

        return new_declaration_line

## PRIVATE

class Counter:
    def __init__(self):
        self.count = 0

    def up(self, how_much=1):
        self.count += how_much


def find_declaration_line(variable, filepath, vault, additional_files=None):
    # Finds line at which the variable is declared
    # WARNING: if the line of the declaration is split, it merges the line in the file
    #          and therefore this function can modify the file
    import re
    # Declaration patterns

    # Standard pattern
    pattern = r"::.*\b%s\b" % variable.name

    # Exception pattern (line is split)
    exception_pattern = r"&.*\b%s\b" % variable.name
    # Common real variable declaration pattern
    real_pattern = r"real.*::"

    def load_lines(filepath):
        with open(filepath, "r") as f:
            # Sometimes people decide that is ok to copy and paste non UTF-8 character in comments. It isn't.
            try:
                lines = [l for l in f]
            except UnicodeDecodeError:
                raise UnicodeDecodeError("The file %s contains non UTF-8 character.\n "
                                         "Please clean, and retry.", filepath)
        return lines

    # Check if the file contains exceptions and fix them
    check_for_exceptions = True
    while check_for_exceptions:
        # Setting this to False, will only activate this again in case a case is found
        check_for_exceptions = False
        lines = load_lines(filepath=filepath)
        for index, line in enumerate(lines):
            line = BasicFunctions.remove_comments(line).strip()
            if current_block.block[index] == variable.procedure:
                m = re.search(exception_pattern, line)
                if m:
                    declaration_displacement = 1
                    if BasicFunctions.remove_comments(lines[index - 1]).strip()[0] == "&":
                        while BasicFunctions.remove_comments(lines[index - declaration_displacement]).strip()[0] == "&":
                            declaration_displacement = declaration_displacement + 1
                    if re.search(real_pattern, lines[index - declaration_displacement], re.I):
                        BasicFunctions.merge_line_range(lines, index - declaration_displacement)
                        check_for_exceptions = True
                        break

    # Lines should be already updated since in case being modified it checks for exceptions again.

    # Search for declarations
    declarations = []

    type_def = None
    for index, line in enumerate(lines):
        line = BasicFunctions.remove_comments(line).strip()
        type_def = _in_which_type_definition(line, type_def)
        if re.search(pattern, line):
            if current_block.block[index] == variable.procedure:
                if variable.is_member_of is None:
                    declarations.append(index)
                elif variable.is_member_of == type_def:
                    declarations.append(index)
                else:
                    pass

    if declarations:
        return declarations

    include_declaration = [line for line in lines if line.count("#include")]
    # If no line has been found in this file searches for possible additional files
    if additional_files is not None:
        for declaration in include_declaration:
            h90_file = declaration.split("#include")[1]
            if h90_file.lower().count("h90"):
                h90_file = h90_file.strip().replace("\"", "")
                additional_files.append(h90_file)
    raise Error.LineNotFound


def find_first_use(variable):
    procedure = variable.procedure
    module = procedure.module

    # pattern = '(^|\W+)%s(\W+|$)' % variable.name
    pattern = r'\W*%s\W*' % variable.name
    searching = False
    loop_level = 0
    loop_start = None
    in_where_statement = False
    where_statement_start = None

    # Scroll all the lines
    for line_index, line in enumerate(module.lines):
        line = line.strip().lower()
        # Until it finds the subroutine to which the variables belongs (routine.name)
        if not searching:
            match = re.search(r"(subroutine|function)\s+\b%s\b" % procedure.name, line, re.IGNORECASE)
            if match:
                searching = True
        # In the subroutine
        else:
            # Scroll all the lines searching for the first use
            if re.search(pattern, line, re.IGNORECASE):
                # skip declarations: bad workaround, but avoids false positive that will block compilation.
                if line.count("::") or line.lower().count("implicit none"):
                    continue
                # Skip the first line
                elif line.count("subroutine"):
                    # Reach the end: the variable has not been used
                    if line.count("end subroutine"):
                        break
                    continue
                # Skip import statements
                elif line.find("use ") == 0:
                    continue
                # Other statement
                elif line.find("data ") == 0:
                    continue
                else:
                    # The variable is found
                    if in_where_statement:
                        # Inside a where statement
                        return module, where_statement_start
                    else:
                        # In a (nested) loop
                        if loop_level > 0:
                            return module, loop_start
                        # On a normal line
                        else:
                            return module, line_index
            elif line.count("end subroutine"):
                break
            # Whenever a 'do loop' starts increment nesting index
            elif re.search(r"^ *do *[\w]* *=", line, re.IGNORECASE):
                if loop_level == 0:
                    loop_start = line_index
                loop_level += 1
            # Whenever a 'do loop' ends decrement nesting index
            # This way match bot the expression 'end do' and 'enddo' but excludes matches like 'enddo loop_name'
            elif re.search(r"^ *end\s*do", line, re.IGNORECASE):
                loop_level -= 1
            # Do the same for where statement
            elif line.find("where") == 0:
                in_where_statement = True
                where_statement_start = line_index
            elif line.find("end where") == 0:
                in_where_statement = False
    return None, None


def _in_which_type_definition(_line, type_def):
    _line = _line.strip().lower()
    _line = BasicFunctions.remove_literals(_line)
    _pattern_type = "^ *type[^(].*::(.*)"
    _pattern_type_alt = "^ *type[^(](.*)"
    _pattern_type_end = "^ *end *type"

    # We are changing block
    if type_def is None:
        m = re.match(_pattern_type, _line, re.I)
        if m:
            type_name = m.group(1).strip()
            return type_name
        n = re.match(_pattern_type_alt, _line, re.I)
        if n:
            type_name = n.group(1).strip()
            return type_name
    else:
        m = re.search(_pattern_type_end, _line, re.I)
        if m:
            return None
    return type_def

